/*
 * block.c
 *
 * The block functions
 *
 * Copyright (C) 2001 Greg Kroah-Hartman <greg@kroah.com>
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stddef.h>	/* for NULL */
#include <stdlib.h>	/* for getenv() */
#include <string.h>	/* for getenv() */
#include <stdio.h>	/* for getenv() */
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/mount.h>
#include <linux/fs.h>
#include <sys/shm.h>
// Volume name library is designed by E.J.
#include "volume.h"
#include "hotplug.h"
#include "System2AP.h"

#define DVDPLAYER_LOCK "/var/lock/.DvdPlayer"
#define HOTPLUG_CONFIG_FILE "/var/lock/hotplug/config"


#define HOTPLUG_MOUNT_TMP "/var/lock/hotplug/mount_tmp"


#define HOTPLUG_MOUNT_TMP_REMOUNT "/tmp/lock/hotplug/mount_tmp"
#define HOTPLUG_CONVERT_TMP "/var/lock/hotplug/convert_tmp"
//#define HOTPLUG_RENAME_TMP "/var/lock/hotplug/rename_tmp"
#define BLOCK_ADD_SIGNAL_NUMBER 48
#define BLOCK_REMOVE_SIGNAL_NUMBER 49
#define COMMAND_TRY 3			// the number of mount try
#define UMOUNT_COMMAND_TRY 3		// the number of umount try
#ifdef CUSTOMER_TV030
#define BLKID_COMMND	"/sbin/blkid"
#endif

// cyhuang (2010/10/28) : Enable new hotplug device naming rule for linux-2.6.34. FIXME : Double-confirn config variable.
//#ifdef LINUX_2_6_34
#define NO_DEVICE_FILESYSTEM		// No use device filesystem and the device file path would be like this: /dev/sda1
//#endif

#define CHECK_RT_SIGNATURE
//#define TRY_MOUNT_MAC_HFS_HFSPLUS

enum device_type {
	TYPE_USB_DEV = 0x0,
	TYPE_SATA_DEV,
	TYPE_CARD_DEV,
	TYPE_NUM,
};

char *type_to_bus_type(enum device_type type)
{
        switch (type) {
        case TYPE_USB_DEV:
                return "USB";
        case TYPE_SATA_DEV:
                return "SATA";
        case TYPE_CARD_DEV:
                return "CARD";

        default:
                return NULL;
        }
}

#ifdef CUSTOMER_TV030
/*
return value:
	Success: 0
	Fail: -1
input: partition name, such as sda1 or sdb1
output: uuid, a given char* space
*/
int get_uuid(char* uuid, const char* partition)
{
	if( access( BLKID_COMMND, F_OK ) == -1 )
	{
		printf("Hotplug: %s() command %s doesn't exist!!\n", __func__, BLKID_COMMND);
		return -1;
	}

	if(uuid == NULL)
	{
		printf("Hotplug: %s() invalid parameter: uuid!!\n", __func__);
		return -1;
	}


	FILE *fp1;
	char buf[512] = {0};
	char found[512] = {0};

	fp1 = popen(BLKID_COMMND, "r");

	while(fgets(buf, 512, fp1)!=0)
	{
		//printf("Hotplug: output: %s\n", buf);  // ummark it for debug

		if(strstr(buf, partition) != NULL) /* find usb disk */
		{
			char* begin = strstr(buf, "UUID=");
			if( begin == NULL )
				break;

			char *string, *token;
			string = strdup(begin+6);	/* parse the string after "UUID=\"" */

			/* find the first non-empty token, delimited by quotation mark */
			while( (token = strsep(&string,"\"")) != NULL )
			{
				printf("   ==> %s (%d)\n",token, strlen(token));
				if (strlen(token) > 4) /* ignore the token length less than 5 */
				{
					memset(found, 0, 512);
					strcpy(found, token);
					printf("                found UUID!!\n");
					break;
				}
			}

		}

		memset(buf, 0, 512);
	}

	pclose(fp1);

	printf("Hotplug: %s() parsed UUID ==> {%s}\n", __func__, found);

	/* check uuid validity */
	if(strlen(found) == 0)	// to do: add more check
	{
		printf("Hotplug: %s() invalid UUID!!\n", __func__);
		return -1;
	}

	strcpy(uuid, found);

	return 0;

}

int get_uuid_remove(char* uuid, const char* partition)
{
	if(uuid == NULL)
	{
		printf("Hotplug: %s() invalid parameter: uuid!!\n", __func__);
		return -1;
	}


	FILE *fp1;
	char buf[512] = {0};
	char found[512] = {0};
	char cmd[512] = {0};

	sprintf(cmd, "cat /proc/mounts | grep %s | awk '{printf $2}'", partition);

	fp1 = popen(cmd, "r");

	if (fgets(buf, 512, fp1) != 0)
	{
		printf("Hotplug: output: %s\n", buf);

		/* find last non-empty token, delimited by quotation mark */
		char *string, *token;
		string = strdup(buf);
		while( (token = strsep(&string,"/")) != NULL )
		{
			printf("   ==> %s\n",token);
			if (strlen(token) > 4)
			{
				memset(found, 0, 512);
				strcpy(found, token);
			}
		}

		memset(buf, 0, 512);
	}
	else
	{
		printf("Hotplug: %s() fail to exec command: {%s}\n", __func__, cmd);
		pclose(fp1);
		return -1;
	}

	pclose(fp1);

	printf("Hotplug: %s() parsed UUID ==> {%s}\n", __func__, found);

	/* check uuid validity */
	if(strlen(found) == 0)	// to do: add more check
	{
		printf("Hotplug: %s() invalid UUID!!\n", __func__);
		return -1;
	}

	strcpy(uuid, found);

	return 0;

}
#endif
char *type_to_device_path(enum device_type type)
{
        switch (type) {
#ifdef SUPPORT_ANDROID_RECOVERY
        case TYPE_USB_DEV:
				#ifdef CUSTOMER_TV030
                return "/mnt/usb";
				#else
				return "/mmnt/udisk";
				#endif
#else
        case TYPE_USB_DEV:
                return "/tmp/usbmounts";
#endif
        case TYPE_SATA_DEV:
                return "/tmp/satamounts";

        case TYPE_CARD_DEV:
#ifdef SUPPORT_ANDROID_RECOVERY
				#ifdef CUSTOMER_TV030
                return "/mnt/sdcard";
				#else
				return "/mmnt/sdcard";
				#endif
#endif
        default:
                return NULL;
        }
}

#ifndef RESCUE_LINUX

// devpath = "/devices/platform/ehci_hcd/usb1/1-1/1-1.1/1-1.1:1.0/host2/target2:0:0/2:0:0:0"
static int get_block_host_num_by_physdevpath(char *host)
{
	char *devpath;
	char *ptr;
	int i = 0;

	printf("#@# %s(%d) getenv (\"PHYSDEVPATH\") %s\n", __func__, __LINE__, getenv ("PHYSDEVPATH"));
	devpath = getenv ("PHYSDEVPATH");

	if(!devpath)
	{
		host = NULL;
		return 1;
	}

	// ptr = "host2/target2:0:0/2:0:0:0"
	ptr = strstr(devpath, "host");
	if(ptr == NULL) {
		// set host0 as default value
		printf("#@# cfyeh-debug %s(%d) NO host_num, set host0 as default value\n", __func__, __LINE__);
		sprintf(host, "host0");
		return 1;
	}

	while(ptr[i] != '/')
	{
		host[i] = ptr[i];
		i++;
	}
	// host = "host2"
	host[i]='\0';

	return 0;
}
#endif

int block_ptp=0;	// If this is a ptp device, the value of block_ptp will be 1.

void do_create_empty_file(const char *name)
{
	int fd;
	if(name) {
		fd = creat(name, 0644);
		if(fd != -1)
			close(fd);
	}
}

/*
   Translate scsi block's device name.
   	Exp: hotplug will give us this: /block/sda/sda1
   	We would add "/sys" in the beginning, remove "sda1", and add "device" in the end: /sys/block/sda/device
   	Then the file links to an informative path: ../../devices/platform/ehci_hcd/usb1/1-1/1-1:1.0/host0/target0:0:0/0:0:0:0
   	Use the last 4 integers and sda'1' to translate into the device path of sda1: /dev/scsi/host0/bus0/target0/lun0/part1
   	Find the sub-string "1-1:1.0". 1-"1":1.0 tells us the device is located in root port. The last integer of 0:0:0:0 tells us the lun number is 0. Therefore, port name would be: 1-0
   	If the sub-string is "1-1.1:1.0". 1-"1.1":1.0 tells us the device is located in port one of a usb hub, and the usb hub is plugged into the root port. The last integer of 0:0:0:0 tells us the lun number is 0. There, port name would be: 1.1-0
 */
#ifndef NO_DEVICE_FILESYSTEM
static int device_name_translating(char *dev, int len1, char *port, int len2)
{
	char *devpath;
//	char num[4];
	char *start, *brk_str, *ptr1=NULL, *ptr2=NULL, *ptr3=NULL, *ptr4=NULL, *ptr5;
	int ret, ret2=0;
	char devices[256];

	if(len1<200 || len2<32)
		return -1;			// Too dangerous
	devpath = getenv ("DEVPATH");

	if(strlen(devpath)>256-10)		// The devpath is too long
		return -1;
	if(!(ptr5 = strrchr(devpath, '/')))
		return -1;
	ptr5++;
	strcpy(devices, "/sys");
	if(!strncmp(ptr5, "mmc", 3) || !strncmp(ptr5, "ms", 2)) {
		if(strstr(ptr5, "p")) {	// ex. mmcblk0p1 msblk0p1
			strncat(devices, devpath, (unsigned int)ptr5-(unsigned int)devpath);
		} else {		// ex. mmcblk0 msblk0
			strcat(devices, devpath);
			strcat(devices, "/");
		}
	} else if(strlen(ptr5) > 3) {
		strncat(devices, devpath, (unsigned int)ptr5-(unsigned int)devpath);
	} else {
		strcat(devices, devpath);
		strcat(devices, "/");
	}
	//printf("[%d] devpath=%s ptr5=%s, devices=%s, dev=%s, port=%s\n",
			//__LINE__, devpath, ptr5, devices, dev, port);
// Some usb disks have no partition. Therefore, sda may need to be mounted, too.
// If sda has partitions, for example, sda1, it will return 1. On the contrary, it will return 0 or -1.
	if(!strstr(ptr5, "sr") &&
			(ptr5[strlen(ptr5)-1]<0x30 || ptr5[strlen(ptr5)-1]>0x39 ||
			!(!strncmp(ptr5, "mmc", 3) && strstr(ptr5, "p")) ||
			!(!strncmp(ptr5, "ms", 2) && strstr(ptr5, "p")))) {
		char partnum_filename[32], read_buffer[2];
		int partnum_fd=0;
		int out_loop=0;
#define WAIT_PART_NUM_READY	200

		sprintf(partnum_filename, "/sys/block/%s/part_num", ptr5);
		while(1) {
			usleep(100000);
			if((partnum_fd=open(partnum_filename, O_RDONLY)) == -1) {
				printf("#######[cfyeh-debug] %s(%d) open %s fail!\n", __func__, __LINE__, partnum_filename);
				break;
			}
			else if(read(partnum_fd, read_buffer, 1) != 1) {
				close(partnum_fd);
				continue;
			} else if(read_buffer[0] == 0x30) {
				printf("Hotplug: \"%s\" has no partition.\n", ptr5);
				ret2 = 0;
				close(partnum_fd);
				break;
			} else if(read_buffer[0]>0x30 && read_buffer[0]<=0x39) {
				printf("Hotplug: \"%s\" has partitions.\n", ptr5);
				ret2 = 1;
				close(partnum_fd);
				break;
			}
			close(partnum_fd);
			out_loop++;
			//printf("Hotplug: out_loop %d\n", out_loop);
			if(out_loop >= WAIT_PART_NUM_READY) {
				printf("Hotplug: wait too long. out_loop %d\n", out_loop);
				break;
			}
		}
	}
	strcat(devices, "device");			// "/sys/block/sda/device" links to, for example, "../../devices/platform/ehci_hcd/usb1/1-1/1-1:1.0/host0/target0:0:0/0:0:0:0". When there are 4 hubs connected in a string, it will be a very long string.
	ret = readlink(devices, devices, len1-1);
	if(ret == -1)
		return -1;
	else
		devices[ret] = 0;

// We would like to know the port that the device is located in
// Now we only support SCSI. For HD, we will copy NULL string to "port".
	if(port && !strncmp(ptr5, "sd", 2)) {
// If the sub-string is "1-1.1:1.0", 1.1 is what we want.
		ptr1 = start = strchr(devices, ':');
		if(!ptr1)
			return -1;
		ret = 0;
		while(*start != '-') {
			start--;
			if(start < devices)
			{
				ret = -1;
				strcpy(port, "");
				break;
			}
		}
		if(!ret)
		{
			start++;
			strncpy(port, start, ptr1-start);
			port[ptr1-start]=0;
// After adding the lun number to the end, port is "1.1-0"
			start = strrchr(devices, ':');
			start++;
			strcat(port, "-");
			strcat(port, start);
			strcat(port, "-");
			if(ptr5[strlen(ptr5)-1]<0x30 || ptr5[strlen(ptr5)-1]>0x39)
				strcat(port, "0");
			else
				strcat(port, &ptr5[strlen(ptr5)-1]);
		}
	} else if(port)
		strcpy(port, "");

// We would like to know the device path of the device
	if(dev) {
		if(!strncmp(ptr5, "sd", 2)) {
// The last 4 integers, for example, 0:0:0:0 represent host, bus, target, and lun respectively.
			start = strrchr(devices, '/');
			start++;
			if((ptr1 = strtok_r(start, ":", &brk_str)))
				if((ptr2 = strtok_r(NULL, ":", &brk_str)))
					if((ptr3 = strtok_r(NULL, ":", &brk_str)))
						ptr4 = strtok_r(NULL, ":", &brk_str);
			if(!ptr4)
				return -1;

			if(strlen(ptr5) > 3)			// ex. sda1
			{
				char *part_ptr;
				for(part_ptr=ptr5+3;(*part_ptr)!=0;part_ptr++) {
					if(isdigit(*part_ptr))
						break;
				}
				sprintf(dev, "/dev/scsi/host%s/bus%s/target%s/lun%s/part%s", ptr1, ptr2, ptr3, ptr4, part_ptr);
			}
			else					// ex. sda
				sprintf(dev, "/dev/scsi/host%s/bus%s/target%s/lun%s/disc", ptr1, ptr2, ptr3, ptr4);
		} else if(!strncmp(ptr5, "sr", 2)) {	//Ken(20081031): for usb CDROM or DVD
		// The last 4 integers, for example, 0:0:0:0 represent host, bus, target, and lun respectively.
			start = strrchr(devices, '/');
			start++;
			if((ptr1 = strtok_r(start, ":", &brk_str)))
				if((ptr2 = strtok_r(NULL, ":", &brk_str)))
					if((ptr3 = strtok_r(NULL, ":", &brk_str)))
						ptr4 = strtok_r(NULL, ":", &brk_str);
			if(!ptr4)
				return -1;

			sprintf(dev, "/dev/scsi/host%s/bus%s/target%s/lun%s/cd", ptr1, ptr2, ptr3, ptr4);
			//printf("[%s] dev=%s\n", __FUNCTION__, dev);
		}else if(!strncmp(ptr5, "hd", 2)) {
			start = strrchr(devices, '/');
			start++;
			if((ptr1 = strtok_r(start, ".", &brk_str)))
				if((ptr2 = strtok_r(NULL, ".", &brk_str)));
			if(!ptr2)
				return -1;
			if(strlen(ptr5) > 3)			// ex. hda1
			{
				char *part_ptr;
				for(part_ptr=ptr5+3;(*part_ptr)!=0;part_ptr++) {
					if(isdigit(*part_ptr))
						break;
				}
				sprintf(dev, "/dev/ide/host%s/bus0/target%s/lun0/part%s", ptr1, ptr2, part_ptr);
			}
			else					// ex. hda
				sprintf(dev, "/dev/ide/host%s/bus0/target%s/lun0/disc", ptr1, ptr2);
		}else if(!strncmp(ptr5, "mmc", 3)) {
			char tmp[64];
			strcpy(tmp, ptr5);
			if((ptr1 = strtok_r(tmp, "k", &brk_str)))
				if((ptr2 = strtok_r(NULL, "p", &brk_str)));
			if(strlen(brk_str))			// ex. mmcblk0p1
				sprintf(dev, "/dev/mmc/blk%s/part%s", ptr2, brk_str);
			else					// ex. mmcblk0
				sprintf(dev, "/dev/mmc/blk%s/disc", ptr2);
		}else if(!strncmp(ptr5, "ms", 2)) {
			char tmp[64];
			strcpy(tmp, ptr5);
			if((ptr1 = strtok_r(tmp, "k", &brk_str)))
				if((ptr2 = strtok_r(NULL, "p", &brk_str)));
			if(strlen(brk_str))			// ex. msblk0p1
				sprintf(dev, "/dev/ms/blk%s/part%s", ptr2, brk_str);
			else					// ex. msblk0
				sprintf(dev, "/dev/ms/blk%s/disc", ptr2);
		}else
			strcpy(dev, "");
	}

	return ret2;
}
#endif

int do_command(char *command, int count, char *lock_file)
{
	int i, ret=1;
	char str_buf[512];
	struct stat st;

	sprintf(str_buf, "%s/lock_%s", HOTPLUG_MOUNT_TMP, lock_file);
	for(i=0;i<count;) {
		if(stat(str_buf, &st) != -1) {
			usleep(200000);
			continue;
		}
		do_create_empty_file(str_buf);
		i++;
		ret=system(command);
		unlink(str_buf);
		if(!WEXITSTATUS(ret) || i>=count)	// For i>=count, this is to jump out as soon as possible
			break;
		usleep(500000);
	}

	printf("Hotplug: %s\tret: %d\n", command, WEXITSTATUS(ret));

	if(ret == -1)			// "ret == -1" means "system" command errors, e.g., fork problem
		return ret;
	else
		return WEXITSTATUS(ret);
}

int do_command2(const char *command, const char *source, const char *target, const char *filesystemtype,
			unsigned long mountflags, const void *data, int count, char *lock_file)
{
	int i, ret=1;
	//char str_buf[512];
	struct stat st;
#ifdef SUPPORT_ANDROID_RECOVERY
//printf("#@#[cfyeh-debug] %s(%d) count %d\n", __func__, __LINE__, count);
	count = 5;
#endif
	//sprintf(str_buf, "%s/lock_%s", HOTPLUG_MOUNT_TMP, lock_file);
	for(i=0;i<count;) {
		//if(stat(str_buf, &st) != -1) {
		//	usleep(200000);
		//	continue;
		//}
		//do_create_empty_file(str_buf);
		i++;
		if ( !strcmp(command, "mount") ){   /* mount function */
#ifdef SUPPORT_ANDROID_RECOVERY
//printf("#@#[cfyeh-debug] %s(%d) source %s\n", __func__, __LINE__, source);
//printf("#@#[cfyeh-debug] %s(%d) target %s\n", __func__, __LINE__, target);
//printf("#@#[cfyeh-debug] %s(%d) filesystemtype %s\n", __func__, __LINE__, filesystemtype);
#endif

 		      ret = mount(source, target, filesystemtype, mountflags, (char *)data);
		     //printf("CALL MOUNT()\n");
		}else{      /* unmount function */
			 ret = umount2(target, mountflags);
			 //printf("CALL UNMOUNT2()\n");
		}
		//unlink(str_buf);
		if( !(ret) || i>=count ) // For i>=count, this is to jump out as soon as possible
			break;
		usleep(500000);
	}

	printf("Hotplug: %s %s %s %s %s 0x%Xret: %d, count %d\n", command, source, target, filesystemtype, data, mountflags, ret, i);
	return ret;
}

#ifdef UFSD_MODULE
int mount_ufsd(const char *name, const char *target, int count, char *ptr)
{
	char mount_args[256];
	int ret;

	sprintf(mount_args, "force,nls=utf8,umask=0,noatime");

	// first, mount device as rw to fix potential error
	ret = do_command2("mount", name, target, "ufsd", MS_SILENT, mount_args, count, ptr);
	if (ret == 0) {
#ifndef HOTPLUG_MOUNT_RW
		ret = do_command2("mount", name, target, "ufsd", MS_REMOUNT | MS_RDONLY, NULL, count, NULL);
#endif
   } else { // if source device is write protected, mount it as read-only
		ret = do_command2("mount", name, target, "ufsd", MS_SILENT | MS_RDONLY, mount_args, count, ptr);
	}

	return ret;
}
#endif /* UFSD_MODULE */

static void check_ready(char* ptr, SYS2AP_MESSAGE sysmsg, int is_all_mounted, char* host_num, enum device_type type, int is_block_ptp)
{
#ifndef RESCUE_LINUX
	if ( strncmp(ptr, "sr", 2) ){
		if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", "Ready_Clear", NULL) > 0)
		{
			printf("#######[cfyeh-debug] %s(%d) find ignore:Ready_Clear at config file.\n", __func__, __LINE__);
			return;
		}
		if(is_all_mounted == 1)
		{
			printf("Hotplug got one BLOCK Hotplug of \"Ready\" from \"%s\"\n", ptr);
			sysmsg.m_type = SYS2AP_MSG_BLOCK;
			if (getenv("PORT_STRUCT"))
				sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_READY,
						(is_block_ptp) ? "ptp" : host_num,
						type_to_bus_type(type), getenv("PORT_STRUCT"));
			else
				sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_READY,
						(is_block_ptp) ? "ptp" : host_num,
						type_to_bus_type(type), ptr);
			RTK_SYS2AP_SendMsg(&sysmsg);

			#ifdef HOTPLUG_FS_MODULIZED
			printf("remove unused filesystem ......\n");
			system("rmmod vfat");
			system("rmmod fat");
			system("rmmod ntfs");
#ifdef UFSD_MODULE
			system("rmmod ufsd");
#endif /* UFSD_MODULE */
			#endif
		}
	}
#endif
}

static void check_clear(char* ptr, SYS2AP_MESSAGE sysmsg, int is_all_umounted, char* host_num, enum device_type type, int is_block_ptp)
{
#ifndef RESCUE_LINUX
	if ( strncmp(ptr, "sr", 2) ){
		if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", "Ready_Clear", NULL) > 0)
		{
			printf("#######[cfyeh-debug] %s(%d) find ignore:Ready_Clear at config file.\n", __func__, __LINE__);
			return;
		}
		if(is_all_umounted == 1)
		{
			printf("Hotplug got one BLOCK Hotplug of \"Clear\" from \"%s\"\n", ptr);
			sysmsg.m_type = SYS2AP_MSG_BLOCK;
			if (getenv("PORT_STRUCT"))
				sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_CLEAR,
						(is_block_ptp) ? "ptp" : host_num,
						type_to_bus_type(type), getenv("PORT_STRUCT"));
			else
				sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_CLEAR,
						(is_block_ptp) ? "ptp" : host_num,
						type_to_bus_type(type), ptr);
			RTK_SYS2AP_SendMsg(&sysmsg);

			#ifdef HOTPLUG_FS_MODULIZED
			printf("remove unused filesystem ......\n");
			system("rmmod vfat");
			system("rmmod fat");
			system("rmmod ntfs");
#ifdef UFSD_MODULE
			system("rmmod ufsd");
#endif /* UFSD_MODULE */
			#endif
		}
	}
#endif
}

#ifndef HOTPLUG_IN_ORDER
inline static void check_ptp_write_lockdata(int is_block_ptp, int fd) {
#ifndef RESCUE_LINUX // we do not mount sata at rescue linux
	if(is_block_ptp)
		write_lockdata_fd(fd, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	else
	{
	printf("#@# %s(%d) getenv (\"PHYSDEVPATH\") %s\n", __func__, __LINE__, getenv ("PHYSDEVPATH"));
		if (getenv("PHYSDEVPATH"))
			write_lockdata_fd(fd, getenv("PHYSDEVPATH"), strlen(getenv("PHYSDEVPATH")));
		else
			write_lockdata_fd(fd, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	}
#endif
}

inline static int check_ptp_testwait_lockdata(int is_block_ptp, const char *filename) {
#ifndef RESCUE_LINUX // we do not mount sata at rescue linux
	if(is_block_ptp)
		return testwait_lockdata(filename, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	else
	{
		printf("#@# %s(%d) getenv (\"PHYSDEVPATH\") %s\n", __func__, __LINE__, getenv ("PHYSDEVPATH"));
		if (getenv("PHYSDEVPATH"))
			return testwait_lockdata(filename, getenv("PHYSDEVPATH"), strlen(getenv("PHYSDEVPATH")));
		else
			return testwait_lockdata(filename, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	}
#endif
}

inline static void check_ptp_testclear_lockdata(int is_block_ptp, const char *filename) {
#ifndef RESCUE_LINUX // we do not mount sata at rescue linux
	if(is_block_ptp)
		testclear_lockdata(filename, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	else
	{
		printf("#@# %s(%d) getenv (\"PHYSDEVPATH\") %s\n", __func__, __LINE__, getenv ("PHYSDEVPATH"));
		if (getenv("PHYSDEVPATH"))
			testclear_lockdata(filename, getenv("PHYSDEVPATH"), strlen(getenv("PHYSDEVPATH")));
		else
			testclear_lockdata(filename, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	}
#endif
}
#endif /* ! HOTPLUG_IN_ORDER */

int echo_command(char *open_filename, char* ptr) {
	int fd;

	if ( (fd = open(open_filename, O_RDWR))==-1 ){
		printf ("block_add(): OPEN %s Error, errno=%s\n", open_filename, strerror(errno));
		return -1;
	}

	if ( write(fd, ptr, strlen(ptr))!= strlen(ptr) ){
		printf ("block_add(): WRITE %s Error, errno=%s\n", open_filename, strerror(errno));
		close(fd);
		return -1;
	}
	close(fd);

	return 0;
}

/*
   An example of hotplug event of usb block device:
      /sbin/hotplug usb
        Environment: PHYSDEVPATH=/devices/platform/ehci_hcd/usb1/1-1/1-1.4/1-1.4:1.0/host7/target7:0:0/7:0:0:0 HOME=/ PATH=/sbin:/bin:/usr/sbin:/usr/bin ACTION=add DEVPATH=/block/sda SUBSYSTEM=usb .....
 */
static int block_add (void)
{
	char *disk_name, disk_name2[32], name[256], name2[64], command[280];
	int ret = 0;
	char *devpath;
	char *ptr;
	char lockfile[256];
	enum device_type type = TYPE_NUM;
	SYS2AP_MESSAGE sysmsg;
	char host_num[8], path[256];
	int label_num = 0;
	char label_path[256];
	int is_all_mounted = 0;
	int fd = -1;
	int zero_size = 0;
#ifdef CHECK_RT_SIGNATURE
	int signature = 0;
	int is_ignore_signature = 0;
#endif /* CHECK_RT_SIGNATURE */
    printf("BUS_TYPE = %s\n",getenv("BUS_TYPE"));

	if(getenv("BUS_TYPE")) {
		if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", getenv("BUS_TYPE"), getenv("PORT_STRUCT")) > 0)
			return 1;
		if(strstr(getenv("BUS_TYPE"), "usb"))
			type = TYPE_USB_DEV;
		else if(strstr(getenv("BUS_TYPE"), "sata"))
			type = TYPE_SATA_DEV;
		else
			type = TYPE_CARD_DEV;
	}
	if(block_ptp) {
		type = TYPE_USB_DEV;
	}

#ifdef RESCUE_LINUX // we do not mount sata at rescue linux
	if(type == TYPE_SATA_DEV)
		return 0;
#endif

#ifdef CHECK_RT_SIGNATURE
	if(getenv("RT_SIGNATURE"))
	if(strstr(getenv("RT_SIGNATURE"), "1")) {
		switch (type) {
		case TYPE_SATA_DEV:
			if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", "signature", getenv("PORT_STRUCT")) > 0) {
				is_ignore_signature = 1;
				signature = 0;
			} else {
				signature = 1;
			}
			break;
		case TYPE_USB_DEV:
			if(parse_config(HOTPLUG_CONFIG_FILE, "notignore", "usbsignature", getenv("PORT_STRUCT")) > 0) {
				signature = 1;
			} else {
				is_ignore_signature = 1;
				signature = 0;
			}
			break;
		default:
			break;
		}
	}
#endif /* CHECK_RT_SIGNATURE */
// The directories in /var/lock/hotplug should be ready or else it may disappear if hotplug comes at the same time with mkdir in rcS
	disk_name=disk_name2;
// cyhuang (2011/05/02) : Chagne mipsel-linux-gcc with new linux-3.4 uclibc header fix stat() bug.
// #ifdef NO_DEVICE_FILESYSTEM
	wait_dir_ready(HOTPLUG_MOUNT_TMP, 100);
#ifndef SUPPORT_ANDROID_RECOVERY
	wait_dir_ready(HOTPLUG_CONVERT_TMP, 100);
#endif
	wait_dir_ready("/sys/block", 100);
// #endif // NO_DEVICE_FILESYSTEM

	devpath = getenv ("DEVPATH");
	if(!(ptr = strrchr(devpath, '/')))	// "/block/sda/sda1"
		return 1;
#ifdef SUPPORT_ANDROID_RECOVERY
//        printf("#@#[cfyeh-debug] %s(%d) devpath %s (pid:%d)\n", __func__, __LINE__, devpath,getpid());
#endif
	// disable mmc block till bootcode pass if mmc is a booting device
	if (strstr(ptr, "mmcblk0"))
		return 1;

#ifndef RESCUE_LINUX
	get_block_host_num_by_physdevpath(host_num); // host_num = "host*"
#endif
	ptr+=1;
	// Only SCSI and PTP block devices will be mounted by block module of hotplug.
	// The names of SCSI block devices will be like sda and sdb1.
	// The names of PTP block devices will be like 1-1:1.0, 1-1.1:1.0, 1-1.1.2:1.0...
	// Here we check for these 2 kinds of block device by comparing the names.
	if(strncmp(ptr, "sd", 2) && strncmp(ptr, "sr", 2) && strncmp(ptr, "mmc", 3) && strncmp(ptr, "ms", 2) &&
			block_ptp == 0) {
		char *ptr1;
		char special_char[4]=":.";
		int special_char_ptr=0, hasnum=0, pattern_ok=1;

		if(!strncmp(ptr, "1-", 2) || !strncmp(ptr, "2-", 2) || !strncmp(ptr, "3-", 2)) {
			for(ptr1=ptr+2;(*ptr1)!=0;ptr1++) {
				if(isdigit(*ptr1)) {
					hasnum=1;
					continue;
				}
				if(hasnum) {
					if(!special_char_ptr && (*ptr1)=='.') {
						hasnum=0;
						continue;
					} else if((*ptr1)==special_char[special_char_ptr]) {
						special_char_ptr++;
						hasnum=0;
						continue;
					}
				}
				pattern_ok=0;
				break;
			}
		}
		if(special_char_ptr != 2 || !hasnum)
			pattern_ok=0;
		if(!pattern_ok) {
			printf("Hotplug: device \"%s\" is skipped.\n", ptr);
			return 1;
		}
	}
	if(getenv("ZERO_SIZE"))
		if(strstr(getenv("ZERO_SIZE"), "1"))
			zero_size = 1;
	if(zero_size) {
		sprintf(lockfile, "%s/.lock_%s_zero", HOTPLUG_MOUNT_TMP, ptr);
		create_lockdata(lockfile);
		printf("\n#@# cfyeh-debug %s(%d) NO Ready signal for ZERO size device\n", __func__, __LINE__);
		return 1;
	}
	sprintf(lockfile, "%s/.lock_%s", HOTPLUG_MOUNT_TMP, ptr);
// Some devices, like mtdblock, won't have "PHYSDEVPATH" environment variable. Therefore, we would like to ignore them as soon as possible.
	if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", ptr, 0) > 0) {
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY

		while((fd = create_lockdata(lockfile)) < 0)
			usleep(10);
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 1;
	}
#ifndef RESCUE_LINUX // we do not mount sata at rescue linux
	// remove sdX lockfile if sdXN is on-going
	if(isdigit(ptr[strlen(ptr)-1])) {
	} else { // for sdX, open sdX and close it, for re-reading partition table
		char read_buffer[2];
		sprintf(command, "/dev/%s", ptr);
		if((fd=open(command, O_RDONLY))) {
			read(fd, read_buffer, 1);
			close(fd);
		} else {
			sprintf(command, "/dev/block/%s", ptr);
			if((fd=open(command, O_RDONLY))) {
				read(fd, read_buffer, 1);
				close(fd);
			}
		}
	}
	// create lock data
	//if(!(signature == 1 && isdigit(ptr[strlen(ptr)-1]) == 0)) // for ! signatued sdx only
	{
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY

		while((fd = create_lockdata(lockfile)) < 0)
			usleep(10);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
	}
#endif
// Delete all the tmp or lock files
	sprintf(command, "%s/lock_%s", HOTPLUG_MOUNT_TMP, ptr);
	remove(command);
	sprintf(command, "%s/.d_%s", HOTPLUG_MOUNT_TMP, ptr);
	remove(command);
	sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);
	remove(command);
	if(block_ptp) {
		sprintf(name, "ptp:%s", devpath);
		setenv("PORT_STRUCT", ptr, 0);
	} else {
// When there is no device filesystem, we still need to know "name2" to parse_config.
#ifndef NO_DEVICE_FILESYSTEM
		ret = device_name_translating(name, 256, name2, 64);
		if(ret < 0) {
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY

			check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
			return ret;
		}
		// for no partition SATA HDD, sleep(1)
		if((ret == 0) && (type == TYPE_SATA_DEV))
			sleep(1);
#else
		// for no partition SATA HDD, sleep(1)
		if(type == TYPE_SATA_DEV)
			sleep(1);

#ifdef SUPPORT_ANDROID_RECOVERY
		sprintf(name, "/dev/block/%s", ptr);
#else
		sprintf(name, "/dev/%s", ptr);
#endif
		if((isdigit(ptr[strlen(ptr)-1]) == 0) || (strstr(ptr, "mmcblk") && !strstr(ptr, "p"))) {// for sdx and mmcblkX only
                            int num;
#ifdef SUPPORT_ANDROID_RECOVERY
                            if (getenv("NPARTS")) {
                                sscanf(getenv("NPARTS"), "%d", &num);

#else
                            if (getenv("PART_NUM")) {
                                sscanf(getenv("PART_NUM"), "%d", &num);
#endif
   //                             printf("%s(line:%d) Enter sdx /mmcblkX check - PART_NUM:%d\n", __FUNCTION__, __LINE__,num);

                                if(num)
                                    ret = 1;
                                else
                                    ret = 0;
                            }
                    }
#endif

#ifndef RESCUE_LINUX
#ifdef CHECK_RT_SIGNATURE
		if(signature == 1)
		{
			if(isdigit(ptr[strlen(ptr)-1]) == 0) // for sdx only
			{
				printf("Hotplug got one BLOCK Hotplug of \"BLOCK_SIGNATURE\" from \"%s\"\n", ptr);
				sysmsg.m_type = SYS2AP_MSG_BLOCK_SIGNATURE;
				sprintf(sysmsg.m_msg, "%s %s", ptr, name);
				RTK_SYS2AP_SendMsg(&sysmsg);
				printf("Hotplug got one BLOCK Hotplug of \"Ready\" from \"%s\"\n", ptr);
				sysmsg.m_type = SYS2AP_MSG_BLOCK;
				if (getenv("PORT_STRUCT"))
					sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_READY,
							(block_ptp) ? "ptp" : host_num,
							type_to_bus_type(type), getenv("PORT_STRUCT"));
				else
					sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_READY,
							(block_ptp) ? "ptp" : host_num,
							type_to_bus_type(type), ptr);
				RTK_SYS2AP_SendMsg(&sysmsg);
			}
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY

			check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
			return 0;
		}
#endif /* CHECK_RT_SIGNATURE */
#endif
#ifndef RESCUE_LINUX
		if ( !strncmp(ptr, "sr", 2) ){	//Ken(20081113): for usb CDROM or DVD
			printf("Hotplug got one USB CDROM Hotplug of \"Add\" from \"%s\"\n", ptr);
			sysmsg.m_type = SYS2AP_MSG_USBCDROM;
			sprintf(sysmsg.m_msg, "%s UP %s", ptr, name);
			//printf("sysmsg.m_msg=%s\n", sysmsg.m_msg);
			RTK_SYS2AP_SendMsg(&sysmsg);
		}
#endif
	}
// "delay:sda1" means sda1 will be blocked here until this instruction is gone.
	while(parse_config(HOTPLUG_CONFIG_FILE, "delay", ptr, 0) > 0) sleep(1);
// "revertdelay:sda1" means only sda1 won't be blocked.
	while(parse_config(HOTPLUG_CONFIG_FILE, "revertdelay", name2, 0) == 0) sleep(1);
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "convert", ptr, 0) > 0) {
		if(strlen(ptr)>3)
			sprintf(command, "%s/%s", HOTPLUG_CONVERT_TMP, ptr+3);
		else
			sprintf(command, "%s/0", HOTPLUG_CONVERT_TMP);
		echo_command(command, name);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 1;
	}
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "convert", name2, 0) == 1) {
		if(strlen(ptr)>3) {
			sprintf(command, "%s/%s", HOTPLUG_CONVERT_TMP, ptr+3);
		} else {
			sprintf(command, "%s/name", HOTPLUG_CONVERT_TMP);
			echo_command(command, ptr);
			sprintf(command, "%s/0", HOTPLUG_CONVERT_TMP);
		}
		echo_command(command, name);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 1;
	}
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "convert/delay", name2, 0) == 1) {
		if(strlen(ptr)>3) {
			sprintf(command, "%s/%s", HOTPLUG_CONVERT_TMP, ptr+3);
		} else {
			sprintf(command, "%s/name", HOTPLUG_CONVERT_TMP);
			echo_command(command, ptr);
			sprintf(command, "%s/0", HOTPLUG_CONVERT_TMP);
		}
		echo_command(command, name);
		while(parse_config(HOTPLUG_CONFIG_FILE, "convert/delay", name2, 0) == 1) sleep(1);
	}
// We parse "ignore" again for delay/revertdelay/convert&delay. After "delay" instruction is removed, we may want to ignore the following processing.
	if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", ptr, 0) > 0) {
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 1;
	}
	if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", name2, 0) == 1) {
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 1;
	}
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "rename", name2, disk_name) == 1) {
		if(strlen(ptr)>3) {
			strcat(disk_name, "_");
			strcat(disk_name, ptr+3);
		}
	} else {
		disk_name = NULL;
	}
// The disc has partition. Therefore, we skip mounting, for example, sda.
	if(!block_ptp && ret == 1) {
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 1;
	}

	sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);

	if((mkdir(command, 0755)) == -1){
		printf("Hotplug: Mkdir \"%s/.%s\" fail.\n", HOTPLUG_MOUNT_TMP, ptr);
		label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		check_ready(ptr, sysmsg, is_all_mounted, host_num, type, block_ptp);
		return 1;
	}
	if(block_ptp) {
		sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);
		if((ret=do_command2("mount", ptr, command, "ptpfs", 0, NULL, COMMAND_TRY, ptr)) == -1) {
			printf("Hotplug: Mount \"%s\" fail.\n", ptr);
			label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
			check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
			check_ready(ptr, sysmsg, is_all_mounted, host_num, type, block_ptp);
			return 1;
		}
	} else {
		printf("[%d]name=%s \n", __LINE__, name);
		if ( !strncmp(ptr, "sr", 2) ){	//Ken(20081031): for usb CDROM or DVD
		}else{
			char mount_args[256];
			unsigned long mountflags = 0;

#ifdef HOTPLUG_MOUNT_RW
			mountflags = MS_SILENT;
#else
			mountflags = MS_SILENT | MS_RDONLY;
#endif /* HOTPLUG_MOUNT_RW */

			sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);

#ifdef HOTPLUG_MOUNT_FAT_FIRST
			sprintf(mount_args, "shortname=winnt,iocharset=utf8,umask=0");
			if((ret=do_command2("mount", name, command, "vfat", mountflags, mount_args, COMMAND_TRY, ptr)) == 0) {
			} else if(ret) {
#ifdef UFSD_MODULE
				system("modprobe ufsd");
				if((ret=mount_ufsd(name, command, 1, ptr)) == 0) {
				} else if(ret) {
#endif /* UFSD_MODULE */
#else /* HOTPLUG_MOUNT_FAT_FIRST */
#ifdef UFSD_MODULE
			system("modprobe ufsd");
			if((ret=mount_ufsd(name, command, COMMAND_TRY, ptr)) == 0) {
			} else if(ret) {
#endif /* UFSD_MODULE */
				sprintf(mount_args, "shortname=winnt,iocharset=utf8,umask=0");
				if((ret=do_command2("mount", name, command, "vfat", mountflags, mount_args, COMMAND_TRY, ptr)) == 0) {
				} else if(ret) {
#endif /* HOTPLUG_MOUNT_FAT_FIRST */
					sprintf(mount_args, "nls=utf8,umask=0");
// Why we still need to retry this command? Because I found that sometimes the first "mount ntfs" will fail alought there is already delay when mounting VFAT.
// I think this is because when "mount sda" and "mount sda1" happen at the same time, it will have problem.
					if((ret=do_command2("mount", name, command, "ntfs", mountflags, mount_args, COMMAND_TRY, ptr)) == 0) {
					} else if(ret) {
						if(!is_ignore_signature && !signature &&
								((ret=do_command2("mount", name, command, "ext3", mountflags, NULL, 1, ptr)) == 0)) {
						} else if (ret) {
							if(!is_ignore_signature && !signature &&
									((ret=do_command2("mount", name, command, "ext4", mountflags, NULL, 1, ptr)) == 0)) {
							}
#ifdef TRY_MOUNT_MAC_HFS_HFSPLUS
							else if(ret) { // try mac filesystem
								if((ret=do_command2("mount", name, command, "hfs", mountflags, NULL, 1, ptr)) == 0) {
								} else if(ret) {
									if((ret=do_command2("mount", name, command, "hfsplus", mountflags, NULL, 1, ptr)) == 0) {
									}
								}
							}
#endif /* TRY_MOUNT_MAC_HFS_HFSPLUS */
						}
					}
				}
#ifdef UFSD_MODULE
			}
#ifndef HOTPLUG_MOUNT_RW
			else {
				ret=do_command2("mount", name, command, "ufsd", MS_REMOUNT | MS_RDONLY, NULL, 1, NULL);
			}
#endif /* ! HOTPLUG_MOUNT_RW */
#endif /* UFSD_MODULE */
		}
	}
	sprintf(name2, "%s/.d_%s", HOTPLUG_MOUNT_TMP, ptr);
	do_create_empty_file(name2);
#ifdef SUPPORT_ANDROID_RECOVERY
	#ifdef CUSTOMER_TV030
	char uuid[512] = {0};
	if (get_uuid(uuid, ptr) == 0)
		sprintf(path, "%s/%s", type_to_device_path(type), uuid);
	else
	#endif
		sprintf(path, "%s/%s", type_to_device_path(type), ptr);
#else
	sprintf(path, "%s/%s", type_to_device_path(TYPE_USB_DEV), ptr);
#endif
	#ifdef CUSTOMER_TV030
	printf("Hotplug: %s %d, path ==> {%s}\n", __func__, __LINE__, path);
	#endif

	if((mkdir(path, 0755)) == -1){
		printf("Hotplug: Mkdir \"%s\" fail.\n", path);
		label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		check_ready(ptr, sysmsg, is_all_mounted, host_num, type, block_ptp);
		return 1;
	}

	sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);
	printf("We do \"%s\" .\n",command);

	if(block_ptp)
		ret = mount(command, path, "ptpfs", MS_MOVE, NULL);
	else
		ret = mount(command, path, NULL, MS_MOVE, NULL);

	if(ret == -1)
	{
		ret = mount(command, path, NULL, MS_MOVE, NULL);
	}

	if(ret == -1)
	{
		ret = mount(command, path, NULL, MS_MOVE, NULL);
	}

	if(ret == -1)
	{
		ret = mount(command, path, NULL, MS_MOVE, NULL);
	}

	if(ret == -1) {
		remove(path);
		remove(name2);
		printf("Hotplug: Mount --move \"%s\" fail.\n", path);
                    printf("wrong with Mount()! %s\n", strerror(errno));
		label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		check_ready(ptr, sysmsg, is_all_mounted, host_num, type, block_ptp);
		return 1;
	}
	if(getenv("EFI_SYSTEM"))
	if(strstr(getenv("EFI_SYSTEM"), "1")) {
		printf("#######[cfyeh-debug] %s(%d) a device w/ EFI_SYSTEM args!\n", __func__, __LINE__);
		umount2(path, MNT_FORCE | MNT_DETACH);
		printf("command: %s\n", command);
		system(command);
		label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		check_ready(ptr, sysmsg, is_all_mounted, host_num, type, block_ptp);
		return 0;
	}

	remove(name2);		// This lock file may be left for a long time.
	if(block_ptp)
		label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
	else
	{
		label_num = add_partition(ptr, NULL, disk_name, &is_all_mounted);
	}

	printf("Hotplug: Mount \"%s\" successfully.\n", ptr);

#ifndef RESCUE_LINUX
	if ( strncmp(ptr, "sr", 2) ){
		if(label_num > 0)
			get_mntpath(label_num, label_path);
		else
			label_path[0]='\0';

		printf("Hotplug got one BLOCK Hotplug of \"Add\" from \"%s\"\n", ptr);
		sysmsg.m_type = SYS2AP_MSG_BLOCK;
		sprintf(sysmsg.m_msg, "%s %s %s %s", ptr, SYS2AP_MSG_UP, path, label_path);
		RTK_SYS2AP_SendMsg(&sysmsg);
	}
#endif
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
	check_ptp_write_lockdata(block_ptp, fd);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
	check_ready(ptr, sysmsg, is_all_mounted, host_num, type, block_ptp);
	return 0;
}

static int block_remove (void)
{
	char name[256], command[280];
	char *devpath;
	char *ptr;
	struct stat st;
	int ret=0;
	char lockfile[256];
	enum device_type type = TYPE_NUM;
	SYS2AP_MESSAGE sysmsg;
	char path[256];
	int is_all_umounted = 0;
	char host_num[8];
#ifdef CHECK_RT_SIGNATURE
	int signature = 0;
#endif /* CHECK_RT_SIGNATURE */

	memset(command, 0x00, 280);
#ifndef RESCUE_LINUX
	get_block_host_num_by_physdevpath(host_num); // host_num = "host*"
#endif
	if(getenv("BUS_TYPE")) {
		if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", getenv("BUS_TYPE"), getenv("PORT_STRUCT")) > 0)
			return 1;
		if(strstr(getenv("BUS_TYPE"), "usb"))
			type = TYPE_USB_DEV;
		else if(strstr(getenv("BUS_TYPE"), "sata"))
			type = TYPE_SATA_DEV;
		else
			type = TYPE_CARD_DEV;
	}
	if(block_ptp) {
		type = TYPE_USB_DEV;
	}

#ifdef CHECK_RT_SIGNATURE
	if(getenv("RT_SIGNATURE"))
	if(strstr(getenv("RT_SIGNATURE"), "1")) {
		switch (type) {
		case TYPE_SATA_DEV:
			if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", "signature", getenv("PORT_STRUCT")) > 0) {
				signature = 0;
			} else {
				signature = 1;
			}
			break;
		case TYPE_USB_DEV:
			if(parse_config(HOTPLUG_CONFIG_FILE, "notignore", "usbsignature", getenv("PORT_STRUCT")) > 0) {
				signature = 1;
			} else {
				signature = 0;
			}
			break;
		default:
			break;
		}
	}
#endif /* CHECK_RT_SIGNATURE */

#ifdef RESCUE_LINUX // we do not mount sata at rescue linux
	if(type == TYPE_SATA_DEV)
		return 0;
#endif

	devpath = getenv ("DEVPATH");
	if(!(ptr = strrchr(devpath, '/')))	// "/block/sda/sda1"
		return 1;

	// disable mmc block till bootcode pass if mmc is a booting device
	if (strstr(ptr, "mmcblk0"))
		return 1;

	ptr+=1;

	// Only SCSI and PTP block devices will be mounted by block module of hotplug.
	// The names of SCSI block devices will be like sda and sdb1.
	// The names of PTP block devices will be like 1-1:1.0, 1-1.1:1.0, 1-1.1.2:1.0...
	// Here we check for these 2 kinds of block device by comparing the names.
	if(strncmp(ptr, "sd", 2) && strncmp(ptr, "sr", 2) && strncmp(ptr, "mmc", 3) && strncmp(ptr, "ms", 2) &&
			block_ptp == 0) {
		char *ptr1;
		char special_char[4]=":.";
		int special_char_ptr=0, hasnum=0, pattern_ok=1;

		if(!strncmp(ptr, "1-", 2) || !strncmp(ptr, "2-", 2) || !strncmp(ptr, "3-", 2)) {
			for(ptr1=ptr+2;(*ptr1)!=0;ptr1++) {
				if(isdigit(*ptr1)) {
					hasnum=1;
					continue;
				}
				if(hasnum) {
					if(!special_char_ptr && (*ptr1)=='.') {
						hasnum=0;
						continue;
					} else if((*ptr1)==special_char[special_char_ptr]) {
						special_char_ptr++;
						hasnum=0;
						continue;
					}
				}
				pattern_ok=0;
				break;
			}
		}
		if(special_char_ptr != 2 || !hasnum)
			pattern_ok=0;
		if(!pattern_ok) {
			printf("Hotplug: device \"%s\" is skipped.\n", ptr);
			return 1;
		}
	}

	sprintf(lockfile, "%s/.lock_%s_zero", HOTPLUG_MOUNT_TMP, ptr);
	if(testclear_lockdata_once(lockfile) == 0) {
		printf("\n#@# cfyeh-debug %s(%d) NO Clear signal for ZERO size device\n", __func__, __LINE__);
		return 1;
	}

	sprintf(lockfile, "%s/.lock_%s", HOTPLUG_MOUNT_TMP, ptr);
	if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", ptr, 0) > 0) {
		printf("#@#[cfyeh-debug] %s(%d)\n", __func__, __LINE__);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_testwait_lockdata(block_ptp, lockfile);
		check_ptp_testclear_lockdata(block_ptp, lockfile);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 1;
	}
	// wait lock file ready
	//if(!(signature == 1 && isdigit(ptr[strlen(ptr)-1]) == 0)) // for ! signatued sdx only
	{
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		if (ENOENT == check_ptp_testwait_lockdata(block_ptp, lockfile)) {
			/*
			For special case, sometimes the hotplug code may receive
			remove but lock file has already been removed by AP.
			In this case, the remove function don't do anything but it
			need to clear share memory
			*/
			int fd_lock;
			int shmid;
			char *pMem = NULL;
			fd_lock = open("/var/lock/hotplug/volume_lock", O_RDONLY);
			if (fd_lock < 0) {
			        MY_PRINTF("Error in open lock file...\n");
				return -1;
			}
			flock(fd_lock, LOCK_EX);

			shmid = shmget(0x23792379, 0x1000, IPC_EXCL | 0660);
			if (shmid == -1) {
				MY_PRINTF("Error in shmget...\n");
				goto ENOENT_out;
			}
			pMem = shmat(shmid, 0, 0);
			if (pMem == (void *)-1) {
				MY_PRINTF("Error in shmat...\n");
				goto ENOENT_out;
			}
			memset(pMem, 0, 0x1000);
			shmdt(pMem);
ENOENT_out:
			fflush(stdout);
			flock(fd_lock, LOCK_UN);
			close(fd_lock);

			return 1;
		}
#endif
#endif /* ! HOTPLUG_IN_ORDER */
	}

#ifndef RESCUE_LINUX
#ifdef CHECK_RT_SIGNATURE
	if(signature == 1)
	{
		if(isdigit(ptr[strlen(ptr)-1]) == 0) // for sdx only
		{
			printf("Hotplug got one BLOCK Hotplug of \"Clear\" from \"%s\"\n", ptr);
			sysmsg.m_type = SYS2AP_MSG_BLOCK;
			if (getenv("PORT_STRUCT"))
				sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_CLEAR,
						(block_ptp) ? "ptp" : host_num,
						type_to_bus_type(type), getenv("PORT_STRUCT"));
			else
				sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_CLEAR,
						(block_ptp) ? "ptp" : host_num,
						type_to_bus_type(type), ptr);
			RTK_SYS2AP_SendMsg(&sysmsg);
		}
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_testclear_lockdata(block_ptp, lockfile);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 0;
	}
#endif /* CHECK_RT_SIGNATURE */
#endif

#ifndef RESCUE_LINUX
	//Ken(20081113): for usb CDROM or DVD
	if ( !strncmp(ptr, "sr", 2) ){
		printf("Hotplug got one USB CDROM Hotplug of \"Remove\" from \"%s\"\n", ptr);
		sysmsg.m_type = SYS2AP_MSG_USBCDROM;
		sprintf(sysmsg.m_msg, "%s DOWN", ptr);
		//printf("sysmsg.m_msg=%s\n", sysmsg.m_msg);
		RTK_SYS2AP_SendMsg(&sysmsg);
	}
#endif

// We cannot use this function at this time because the sys directory of the device may disappear immediately when it is unplugged.
// However, we can use this on card reader which always has sys directory.
// HOTPLUG_RENAME_TMP directory stores the files that have been renamed. Use them instead of "device_name_translating"
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "convert", ptr, 0) > 0) {
		if(strlen(ptr)>3) {
			sprintf(command, "%s/%s", HOTPLUG_CONVERT_TMP, ptr+3);
			printf("Hotplug: Delete convert file \"%s/%s\".\n", HOTPLUG_CONVERT_TMP, ptr+3);
		} else {
			sprintf(command, "%s/0", HOTPLUG_CONVERT_TMP);
			printf("Hotplug: Delete convert file \"%s/0\".\n", HOTPLUG_CONVERT_TMP);
		}
		remove(command);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_testclear_lockdata(block_ptp, lockfile);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		return 1;
	}

	// test sata path to see if the device is from sata
#ifdef SUPPORT_ANDROID_RECOVERY
	#ifdef CUSTOMER_TV030
	char uuid[512] = {0};
	if (get_uuid_remove(uuid, ptr) == 0)
		sprintf(path, "%s/%s", type_to_device_path(type), uuid);
	else
	#endif
		sprintf(path, "%s/%s", type_to_device_path(type), ptr);
#else
	sprintf(path, "%s/%s", type_to_device_path(TYPE_USB_DEV), ptr);
#endif

// cyhuang (2010/10/28) : Disable for linux-2.6.34. FIXME : find stat() function's bug.
#ifdef NO_DEVICE_FILESYSTEM
	if(stat(path, &st) || !S_ISDIR(st.st_mode)) {
		printf("Hotplug: Cannot find \"%s\".\n", path);
		del_partition(ptr, &is_all_umounted);
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
		check_ptp_testclear_lockdata(block_ptp, lockfile);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
		check_clear(ptr, sysmsg, is_all_umounted, host_num, type, block_ptp);
		return 1;
	}
#endif

	// Lock file to notify that the corresponding directory cannot be accessed
	sprintf(name, "%s/.d_%s", HOTPLUG_MOUNT_TMP, ptr);
	do_create_empty_file(name);

	del_partition(ptr, &is_all_umounted);

#ifndef RESCUE_LINUX
	if ( strncmp(ptr, "sr", 2) ){
		printf("Hotplug got one BLOCK Hotplug of \"Remove\" from \"%s\"\n", ptr);
		sysmsg.m_type = SYS2AP_MSG_BLOCK;
		sprintf(sysmsg.m_msg, "%s %s", ptr, SYS2AP_MSG_DOWN);
		RTK_SYS2AP_SendMsg(&sysmsg);
	}
#endif

	umount2(path, MNT_FORCE | MNT_DETACH);

	if(getenv("EFI_SYSTEM"))
	if(strstr(getenv("EFI_SYSTEM"), "1"))
		goto out;

	if(do_command(command, UMOUNT_COMMAND_TRY, ptr))
		ret = 1;
	else {
		remove(path);
		remove(name);
	}
	printf("Hotplug: Umount %s successfully.\n", ptr);
out:
#ifndef HOTPLUG_IN_ORDER
#ifndef SUPPORT_ANDROID_RECOVERY
	check_ptp_testclear_lockdata(block_ptp, lockfile);
#endif
#endif /* ! HOTPLUG_IN_ORDER */
	check_clear(ptr, sysmsg, is_all_umounted, host_num, type, block_ptp);
	return ret;
}

static int block_change (void)
{
//static int block_remove (void)
{
	char name[256], command[280];
	char *devpath;
	char *ptr;
	struct stat st;
	int ret=0;
	enum device_type type = TYPE_NUM;
	SYS2AP_MESSAGE sysmsg;
	char path[256];
	int is_all_umounted = 0;
	char host_num[8];
#ifdef CHECK_RT_SIGNATURE
	int signature = 0;
#endif /* CHECK_RT_SIGNATURE */

#ifndef RESCUE_LINUX
	get_block_host_num_by_physdevpath(host_num); // host_num = "host*"
#endif
	if(getenv("BUS_TYPE")) {
		if(strstr(getenv("BUS_TYPE"), "usb"))
			type = TYPE_USB_DEV;
		else if(strstr(getenv("BUS_TYPE"), "sata"))
			type = TYPE_SATA_DEV;
		else
			type = TYPE_CARD_DEV;
	}
	if(block_ptp) {
		type = TYPE_USB_DEV;
	}

#ifdef CHECK_RT_SIGNATURE
	if(getenv("RT_SIGNATURE"))
	if(strstr(getenv("RT_SIGNATURE"), "1")) {
		switch (type) {
		case TYPE_SATA_DEV:
			if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", "signature", getenv("PORT_STRUCT")) > 0) {
				signature = 0;
			} else {
				signature = 1;
			}
			break;
		case TYPE_USB_DEV:
			if(parse_config(HOTPLUG_CONFIG_FILE, "notignore", "usbsignature", getenv("PORT_STRUCT")) > 0) {
				signature = 1;
			} else {
				signature = 0;
			}
			break;
		default:
			break;
		}
	}
#endif /* CHECK_RT_SIGNATURE */

#ifdef RESCUE_LINUX // we do not mount sata at rescue linux
	if(type == TYPE_SATA_DEV)
		return 0;
#endif

	devpath = getenv ("DEVPATH");
	if(!(ptr = strrchr(devpath, '/')))	// "/block/sda/sda1"
		return 1;

	// disable mmc block till bootcode pass if mmc is a booting device
	if (strstr(ptr, "mmc"))
		return 1;

	ptr+=1;

	// Only SCSI and PTP block devices will be mounted by block module of hotplug.
	// The names of SCSI block devices will be like sda and sdb1.
	// The names of PTP block devices will be like 1-1:1.0, 1-1.1:1.0, 1-1.1.2:1.0...
	// Here we check for these 2 kinds of block device by comparing the names.
	if(strncmp(ptr, "sd", 2) && strncmp(ptr, "sr", 2) && strncmp(ptr, "mmc", 3) && strncmp(ptr, "ms", 2) &&
			block_ptp == 0) {
		char *ptr1;
		char special_char[4]=":.";
		int special_char_ptr=0, hasnum=0, pattern_ok=1;

		if(!strncmp(ptr, "1-", 2) || !strncmp(ptr, "2-", 2) || !strncmp(ptr, "3-", 2)) {
			for(ptr1=ptr+2;(*ptr1)!=0;ptr1++) {
				if(isdigit(*ptr1)) {
					hasnum=1;
					continue;
				}
				if(hasnum) {
					if(!special_char_ptr && (*ptr1)=='.') {
						hasnum=0;
						continue;
					} else if((*ptr1)==special_char[special_char_ptr]) {
						special_char_ptr++;
						hasnum=0;
						continue;
					}
				}
				pattern_ok=0;
				break;
			}
		}
		if(special_char_ptr != 2 || !hasnum)
			pattern_ok=0;
		if(!pattern_ok) {
			printf("Hotplug: device \"%s\" is skipped.\n", ptr);
			return 1;
		}
	}

#ifndef RESCUE_LINUX
#ifdef CHECK_RT_SIGNATURE
	if(signature == 1)
	{
		if(isdigit(ptr[strlen(ptr)-1]) == 0) // for sdx only
		{
			printf("Hotplug got one BLOCK Hotplug of \"Clear\" from \"%s\"\n", ptr);
			sysmsg.m_type = SYS2AP_MSG_BLOCK;
			if (getenv("PORT_STRUCT"))
				sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_CLEAR,
						(block_ptp) ? "ptp" : host_num,
						type_to_bus_type(type), getenv("PORT_STRUCT"));
			else
				sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_CLEAR,
						(block_ptp) ? "ptp" : host_num,
						type_to_bus_type(type), ptr);
			RTK_SYS2AP_SendMsg(&sysmsg);
		}
		return 0;
	}
#endif /* CHECK_RT_SIGNATURE */
#endif

#ifndef RESCUE_LINUX
	//Ken(20081113): for usb CDROM or DVD
	if ( !strncmp(ptr, "sr", 2) ){
		printf("Hotplug got one USB CDROM Hotplug of \"Remove\" from \"%s\"\n", ptr);
		sysmsg.m_type = SYS2AP_MSG_USBCDROM;
		sprintf(sysmsg.m_msg, "%s DOWN", ptr);
		//printf("sysmsg.m_msg=%s\n", sysmsg.m_msg);
		RTK_SYS2AP_SendMsg(&sysmsg);
	}
#endif

// We cannot use this function at this time because the sys directory of the device may disappear immediately when it is unplugged.
// However, we can use this on card reader which always has sys directory.
// HOTPLUG_RENAME_TMP directory stores the files that have been renamed. Use them instead of "device_name_translating"
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "convert", ptr, 0) > 0) {
		if(strlen(ptr)>3) {
			sprintf(command, "%s/%s", HOTPLUG_CONVERT_TMP, ptr+3);
			printf("Hotplug: Delete convert file \"%s/%s\".\n", HOTPLUG_CONVERT_TMP, ptr+3);
		} else {
			sprintf(command, "%s/0", HOTPLUG_CONVERT_TMP);
			printf("Hotplug: Delete convert file \"%s/0\".\n", HOTPLUG_CONVERT_TMP);
		}
		remove(command);
		return 1;
	}
	// Check if the directory exists.

	// test sata path to see if the device is from sata
	sprintf(path, "%s/%s", type_to_device_path(TYPE_USB_DEV), ptr);

// cyhuang (2010/10/28) : Disable for linux-2.6.34. FIXME : find stat() function's bug.
#ifdef NO_DEVICE_FILESYSTEM
	if(stat(path, &st) || !S_ISDIR(st.st_mode)) {
		printf("Hotplug: Cannot find \"%s\".\n", path);
		del_partition(ptr, &is_all_umounted);
		return 1;
	}
#endif

	// Lock file to notify that the corresponding directory cannot be accessed
	sprintf(name, "%s/.d_%s", HOTPLUG_MOUNT_TMP, ptr);
	do_create_empty_file(name);

	del_partition(ptr, &is_all_umounted);

#ifndef RESCUE_LINUX
	if ( strncmp(ptr, "sr", 2) ){
		printf("Hotplug got one BLOCK Hotplug of \"Remove\" from \"%s\"\n", ptr);
		sysmsg.m_type = SYS2AP_MSG_BLOCK;
		sprintf(sysmsg.m_msg, "%s %s", ptr, SYS2AP_MSG_DOWN);
		RTK_SYS2AP_SendMsg(&sysmsg);
	}
#endif

	umount2(path, MNT_FORCE | MNT_DETACH);

	if(getenv("EFI_SYSTEM"))
	if(strstr(getenv("EFI_SYSTEM"), "1"))
		goto out;

	if(do_command(command, UMOUNT_COMMAND_TRY, ptr))
		ret = 1;
	printf("Hotplug: Umount %s successfully.\n", ptr);
out:
	return ret;
}
//static int block_add (void)
{
	char *disk_name, disk_name2[32], name[256], name2[64], command[280];
	int ret = 0;
	char *devpath;
	char *ptr;
	enum device_type type = TYPE_NUM;
	SYS2AP_MESSAGE sysmsg;
	char host_num[8], path[256];
	int label_num = 0;
	char label_path[256];
	int is_all_mounted = 0;
#ifdef CHECK_RT_SIGNATURE
	int signature = 0;
	int is_ignore_signature = 0;
#endif /* CHECK_RT_SIGNATURE */

    printf("BUS_TYPE = %s\n",getenv("BUS_TYPE"));

	if(getenv("BUS_TYPE")) {
		if(strstr(getenv("BUS_TYPE"), "usb"))
			type = TYPE_USB_DEV;
		else if(strstr(getenv("BUS_TYPE"), "sata"))
			type = TYPE_SATA_DEV;
		else
			type = TYPE_CARD_DEV;
	}
	if(block_ptp) {
		type = TYPE_USB_DEV;
	}

#ifdef RESCUE_LINUX // we do not mount sata at rescue linux
	if(type == TYPE_SATA_DEV)
		return 0;
#endif

#ifdef CHECK_RT_SIGNATURE
	if(getenv("RT_SIGNATURE"))
	if(strstr(getenv("RT_SIGNATURE"), "1")) {
		switch (type) {
		case TYPE_SATA_DEV:
			if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", "signature", getenv("PORT_STRUCT")) > 0) {
				is_ignore_signature = 1;
				signature = 0;
			} else {
				signature = 1;
			}
			break;
		case TYPE_USB_DEV:
			if(parse_config(HOTPLUG_CONFIG_FILE, "notignore", "usbsignature", getenv("PORT_STRUCT")) > 0) {
				signature = 1;
			} else {
				is_ignore_signature = 1;
				signature = 0;
			}
			break;
		default:
			break;
		}
	}
#endif /* CHECK_RT_SIGNATURE */

// The directories in /var/lock/hotplug should be ready or else it may disappear if hotplug comes at the same time with mkdir in rcS
	disk_name=disk_name2;

// cyhuang (2011/05/02) : Chagne mipsel-linux-gcc with new linux-3.4 uclibc header fix stat() bug.
// #ifdef NO_DEVICE_FILESYSTEM
	wait_dir_ready(HOTPLUG_MOUNT_TMP, 100);

#ifndef SUPPORT_ANDROID_RECOVERY
	wait_dir_ready(HOTPLUG_CONVERT_TMP, 100);
#endif
	wait_dir_ready("/sys/block", 100);

// #endif // NO_DEVICE_FILESYSTEM

	devpath = getenv ("DEVPATH");
	if(!(ptr = strrchr(devpath, '/')))	// "/block/sda/sda1"
		return 1;

	// disable mmc block till bootcode pass if mmc is a booting device
	if (strstr(ptr, "mmc"))
		return 1;

#ifndef RESCUE_LINUX
	get_block_host_num_by_physdevpath(host_num); // host_num = "host*"
#endif
	ptr+=1;

	// Only SCSI and PTP block devices will be mounted by block module of hotplug.
	// The names of SCSI block devices will be like sda and sdb1.
	// The names of PTP block devices will be like 1-1:1.0, 1-1.1:1.0, 1-1.1.2:1.0...
	// Here we check for these 2 kinds of block device by comparing the names.
	if(strncmp(ptr, "sd", 2) && strncmp(ptr, "sr", 2) && strncmp(ptr, "mmc", 3) && strncmp(ptr, "ms", 2) &&
			block_ptp == 0) {
		char *ptr1;
		char special_char[4]=":.";
		int special_char_ptr=0, hasnum=0, pattern_ok=1;

		if(!strncmp(ptr, "1-", 2) || !strncmp(ptr, "2-", 2) || !strncmp(ptr, "3-", 2)) {
			for(ptr1=ptr+2;(*ptr1)!=0;ptr1++) {
				if(isdigit(*ptr1)) {
					hasnum=1;
					continue;
				}
				if(hasnum) {
					if(!special_char_ptr && (*ptr1)=='.') {
						hasnum=0;
						continue;
					} else if((*ptr1)==special_char[special_char_ptr]) {
						special_char_ptr++;
						hasnum=0;
						continue;
					}
				}
				pattern_ok=0;
				break;
			}
		}
		if(special_char_ptr != 2 || !hasnum)
			pattern_ok=0;
		if(!pattern_ok) {
			printf("Hotplug: device \"%s\" is skipped.\n", ptr);
			return 1;
		}
	}

// Delete all the tmp or lock files
	if(block_ptp) {
		sprintf(name, "ptp:%s", devpath);
		setenv("PORT_STRUCT", ptr, 0);
	} else {
// When there is no device filesystem, we still need to know "name2" to parse_config.
#ifndef NO_DEVICE_FILESYSTEM
		ret = device_name_translating(name, 256, name2, 64);
		if(ret < 0) {
			return ret;
		}
		// for no partition SATA HDD, sleep(1)
		if((ret == 0) && (type == TYPE_SATA_DEV))
			sleep(1);
#else
		// for no partition SATA HDD, sleep(1)
		if(type == TYPE_SATA_DEV)
			sleep(1);

		sprintf(name, "/dev/block/%s", ptr);

		if((isdigit(ptr[strlen(ptr)-1]) == 0) || (strstr(ptr, "mmcblk") && !strstr(ptr, "p"))) {// for sdx and mmcblkX only
                            int num;

#ifdef SUPPORT_ANDROID_RECOVERY
                            if (getenv("NPARTS")) {
                                sscanf(getenv("NPARTS"), "%d", &num);

#else
                            if (getenv("PART_NUM")) {
                                sscanf(getenv("PART_NUM"), "%d", &num);
#endif
                                MY_PRINTF("(%s:line%d) Enter sdx /mmcblkX check part number: %d (env)\n", __FUNCTION__, __LINE__,num);
                                if(num)
                                    ret = 1;
                                else
                                    ret = 0;
                            }
                    }
#endif

#ifndef RESCUE_LINUX
#ifdef CHECK_RT_SIGNATURE
		if(signature == 1)
		{
			if(isdigit(ptr[strlen(ptr)-1]) == 0) // for sdx only
			{
				printf("Hotplug got one BLOCK Hotplug of \"BLOCK_SIGNATURE\" from \"%s\"\n", ptr);
				sysmsg.m_type = SYS2AP_MSG_BLOCK_SIGNATURE;
				sprintf(sysmsg.m_msg, "%s %s", ptr, name);
				RTK_SYS2AP_SendMsg(&sysmsg);
				printf("Hotplug got one BLOCK Hotplug of \"Ready\" from \"%s\"\n", ptr);
				sysmsg.m_type = SYS2AP_MSG_BLOCK;
				if (getenv("PORT_STRUCT"))
					sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_READY,
							(block_ptp) ? "ptp" : host_num,
							type_to_bus_type(type), getenv("PORT_STRUCT"));
				else
					sprintf(sysmsg.m_msg, "%s %s %s %s %s", ptr, SYS2AP_MSG_READY,
							(block_ptp) ? "ptp" : host_num,
							type_to_bus_type(type), ptr);
				RTK_SYS2AP_SendMsg(&sysmsg);
			}
			return 0;
		}
#endif /* CHECK_RT_SIGNATURE */
#endif

#ifndef RESCUE_LINUX
		if ( !strncmp(ptr, "sr", 2) ){	//Ken(20081113): for usb CDROM or DVD
			printf("Hotplug got one USB CDROM Hotplug of \"Add\" from \"%s\"\n", ptr);
			sysmsg.m_type = SYS2AP_MSG_USBCDROM;
			sprintf(sysmsg.m_msg, "%s UP %s", ptr, name);
			//printf("sysmsg.m_msg=%s\n", sysmsg.m_msg);
			RTK_SYS2AP_SendMsg(&sysmsg);
		}
#endif
	}
// "delay:sda1" means sda1 will be blocked here until this instruction is gone.
	while(parse_config(HOTPLUG_CONFIG_FILE, "delay", ptr, 0) > 0) sleep(1);
// "revertdelay:sda1" means only sda1 won't be blocked.
	while(parse_config(HOTPLUG_CONFIG_FILE, "revertdelay", name2, 0) == 0) sleep(1);
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "convert", ptr, 0) > 0) {
		if(strlen(ptr)>3)
			sprintf(command, "%s/%s", HOTPLUG_CONVERT_TMP, ptr+3);
		else
			sprintf(command, "%s/0", HOTPLUG_CONVERT_TMP);
		echo_command(command, name);
		return 1;
	}
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "convert", name2, 0) == 1) {
		if(strlen(ptr)>3) {
			sprintf(command, "%s/%s", HOTPLUG_CONVERT_TMP, ptr+3);
		} else {
			sprintf(command, "%s/name", HOTPLUG_CONVERT_TMP);
			echo_command(command, ptr);
			sprintf(command, "%s/0", HOTPLUG_CONVERT_TMP);
		}
		echo_command(command, name);
		return 1;
	}
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "convert/delay", name2, 0) == 1) {
		if(strlen(ptr)>3) {
			sprintf(command, "%s/%s", HOTPLUG_CONVERT_TMP, ptr+3);
		} else {
			sprintf(command, "%s/name", HOTPLUG_CONVERT_TMP);
			echo_command(command, ptr);
			sprintf(command, "%s/0", HOTPLUG_CONVERT_TMP);
		}
		echo_command(command, name);
		while(parse_config(HOTPLUG_CONFIG_FILE, "convert/delay", name2, 0) == 1) sleep(1);
	}
// We parse "ignore" again for delay/revertdelay/convert&delay. After "delay" instruction is removed, we may want to ignore the following processing.
	if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", ptr, 0) > 0) {
		return 1;
	}
	if(parse_config(HOTPLUG_CONFIG_FILE, "ignore", name2, 0) == 1) {
		return 1;
	}
	if(!block_ptp && parse_config(HOTPLUG_CONFIG_FILE, "rename", name2, disk_name) == 1) {
		if(strlen(ptr)>3) {
			strcat(disk_name, "_");
			strcat(disk_name, ptr+3);
		}
	} else {
		disk_name = NULL;
	}
// The disc has partition. Therefore, we skip mounting, for example, sda.
	if(!block_ptp && ret == 1) {
		return 1;
	}
	sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);
	if((mkdir(command, 0755)) == -1){
		printf("Hotplug: Mkdir \"%s/.%s\" fail.\n", HOTPLUG_MOUNT_TMP, ptr);
		label_num = add_partition(ptr, name, disk_name, 0);
		return 1;
	}
	if(block_ptp) {
		sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);
		if((ret=do_command2("mount", ptr, command, "ptpfs", 0, NULL, COMMAND_TRY, ptr)) == -1) {
			printf("Hotplug: Mount \"%s\" fail.\n", ptr);
			label_num = add_partition(ptr, name, disk_name, 0);
			return 1;
		}
	} else {
		//printf("[%d]name=%s \n", __LINE__, name);
		if ( !strncmp(ptr, "sr", 2) ){	//Ken(20081031): for usb CDROM or DVD
		} else {
			char mount_args[256];
			unsigned long mountflags = 0;

#ifdef HOTPLUG_MOUNT_RW
			mountflags = MS_SILENT;
#else
			mountflags = MS_SILENT | MS_RDONLY;
#endif /* HOTPLUG_MOUNT_RW */

			sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);

#ifdef UFSD_MODULE
			if((ret=mount_ufsd(name, command, COMMAND_TRY, ptr)) == 0) {
			} else if(ret) {
#endif /* UFSD_MODULE */
				sprintf(mount_args, "shortname=winnt,iocharset=utf8,umask=0");
				if((ret=do_command2("mount", name, command, "vfat", mountflags, mount_args, COMMAND_TRY, ptr)) == 0) {
				} else if(ret) {
					sprintf(mount_args, "iocharset=utf8,umask=0");
// Why we still need to retry this command? Because I found that sometimes the first "mount ntfs" will fail alought there is already delay when mounting VFAT.
// I think this is because when "mount sda" and "mount sda1" happen at the same time, it will have problem.
					if((ret=do_command2("mount", name, command, "ntfs", mountflags, mount_args, 1, ptr)) == 0) {
					} else if(ret) {
						if(!is_ignore_signature && !signature &&
								((ret=do_command2("mount", name, command, "ext3", mountflags, NULL, 1, ptr)) == 0)) {
						} else if (ret) {
							if(!is_ignore_signature && !signature &&
									((ret=do_command2("mount", name, command, "ext4", mountflags, NULL, 1, ptr)) == 0)) {
							}
#ifdef TRY_MOUNT_MAC_HFS_HFSPLUS
							else if(ret) { // try mac filesystem
								if((ret=do_command2("mount", name, command, "hfs", mountflags, NULL, 1, ptr)) == 0) {
								} else if(ret) {
									if((ret=do_command2("mount", name, command, "hfsplus", mountflags, NULL, 1, ptr)) == 0) {
									}
								}
							}
#endif /* TRY_MOUNT_MAC_HFS_HFSPLUS */
						}
					}
				}
#ifdef UFSD_MODULE
			}
#ifndef HOTPLUG_MOUNT_RW
			else {
				ret=do_command2("mount", name, command, "ufsd", MS_REMOUNT | MS_RDONLY, NULL, 1, NULL);
			}
#endif /* ! HOTPLUG_MOUNT_RW */
#endif /* UFSD_MODULE */
		}
	}

	sprintf(name2, "%s/.d_%s", HOTPLUG_MOUNT_TMP, ptr);
	do_create_empty_file(name2);
	sprintf(path, "%s/%s", type_to_device_path(TYPE_USB_DEV), ptr);

	if((mkdir(path, 0755)) == -1){
		printf("Hotplug: Mkdir \"%s\" fail.\n", path);
		label_num = add_partition(ptr, name, disk_name, 0);
		return 1;
	}

	sprintf(command, "%s/.%s", HOTPLUG_MOUNT_TMP, ptr);
	if(block_ptp)
		ret = mount(command, path, "ptpfs", MS_MOVE, NULL);
	else
		ret = mount(command, path, NULL, MS_MOVE, NULL);
	if(ret == -1) {
		remove(path);
		remove(name2);
		printf("Hotplug: Mount --move \"%s\" fail.\n", path);
		label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
		return 1;
	}

	if(getenv("EFI_SYSTEM"))
	if(strstr(getenv("EFI_SYSTEM"), "1")) {
		printf("#######[cfyeh-debug] %s(%d) a device w/ EFI_SYSTEM args!\n", __func__, __LINE__);
		umount2(path, MNT_FORCE | MNT_DETACH);
		printf("command: %s\n", command);
		system(command);
		label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
		return 0;
	}

	remove(name2);		// This lock file may be left for a long time.
	if(block_ptp)
		label_num = add_partition(ptr, name, disk_name, &is_all_mounted);
	else
	{
		label_num = add_partition(ptr, NULL, disk_name, &is_all_mounted);
	}

	printf("Hotplug: Mount \"%s\" successfully.\n", ptr);

#ifndef RESCUE_LINUX
	if ( strncmp(ptr, "sr", 2) ){
		if(label_num > 0)
			get_mntpath(label_num, label_path);
		else
			label_path[0]='\0';

		printf("Hotplug got one BLOCK Hotplug of \"Add\" from \"%s\"\n", ptr);
		sysmsg.m_type = SYS2AP_MSG_BLOCK;
		sprintf(sysmsg.m_msg, "%s %s %s %s", ptr, SYS2AP_MSG_UP, path, label_path);
		RTK_SYS2AP_SendMsg(&sysmsg);
	}
#endif

	return 0;
}
}

static struct subsystem block_subsystem[] = {
	{ ADD_STRING, block_add },
	{ REMOVE_STRING, block_remove },
//	{ CHANGE_STRING, block_change },
	{ NULL, NULL }
};


int block_handler (void)
{
	char * action;
	int ret = 0;
#ifdef HOTPLUG_IN_ORDER
	char lockfile[256];
	int fd;
#endif /* HOTPLUG_IN_ORDER */

	action = getenv ("ACTION");
	dbg ("action = %s", action);
	if (action == NULL) {
		dbg ("missing ACTION environment variable, aborting.");
		return 1;
	}

#ifdef HOTPLUG_IN_ORDER
#ifndef RESCUE_LINUX // we do not mount sata at rescue linux
	// create lock data
	sprintf(lockfile, "%s/.lock_block", HOTPLUG_MOUNT_TMP);
	while((fd = create_lockdata(lockfile)) < 0)
		usleep(500000);
	if(block_ptp)
		write_lockdata_fd(fd, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	else
	{
		printf("#@# %s(%d) getenv (\"PHYSDEVPATH\") %s\n", __func__, __LINE__, getenv ("PHYSDEVPATH"));
		if (getenv("PHYSDEVPATH"))
			write_lockdata_fd(fd, getenv("PHYSDEVPATH"), strlen(getenv("PHYSDEVPATH")));
		else
			write_lockdata_fd(fd, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	}
	close(fd);
#endif
#endif /* HOTPLUG_IN_ORDER */

	ret = call_subsystem (action, block_subsystem);

#ifdef HOTPLUG_IN_ORDER
#ifndef RESCUE_LINUX // we do not mount sata at rescue linux
	if(block_ptp)
		testclear_lockdata(lockfile, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	else
	{
		printf("#@# %s(%d) getenv (\"PHYSDEVPATH\") %s\n", __func__, __LINE__, getenv ("PHYSDEVPATH"));
		if (getenv("PHYSDEVPATH"))
			testclear_lockdata(lockfile, getenv("PHYSDEVPATH"), strlen(getenv("PHYSDEVPATH")));
		else
			testclear_lockdata(lockfile, getenv("DEVPATH"), strlen(getenv("DEVPATH")));
	}
#endif
#endif /* HOTPLUG_IN_ORDER */

	return ret;

}


